@@ -85,6 +85,7 @@ let trace = set_trace
 
 let usage =
   Trace_ppx_runtime.Runtime.usage
+  type warning_id = Util.warning_id = LinearVariable | UndeclaredGlobal | FlexClause | ImplicationPrecedence
 
 let set_warn = Util.set_warn
 let set_error = Util.set_error
@@ -109,7 +110,7 @@ module Ast = struct
   module Name = struct
     include Ast.Func
     type constant = int
-    let is_global f i = show f = Util.Constants.Map.find i Data.Global_symbols.table.c2s
+    let is_global f i = equal f (Util.Constants.Map.find i Data.Global_symbols.table.c2s |> Data.Symbol.get_func)
   end
   module Opaque = Util.CData
 end
@@ -181,6 +182,7 @@ module Compile = struct
     Compiler.query_of_ast s_p t (fun st -> st)
 
   let total_type_checking_time q = Compiler.total_type_checking_time q
+  let total_det_checking_time q = Compiler.total_det_checking_time q
 
   module StrSet = Util.StrSet
 
@@ -188,6 +190,7 @@ module Compile = struct
     defined_variables : StrSet.t;
     print_units : bool;
     time_typechecking : bool;
+    skip_det_checking: bool;
   }
   let default_flags = Compiler.default_flags
   let optimize = Compiler.optimize_query
@@ -200,6 +203,14 @@ module Compile = struct
   let signature u = Compiler.signature_of_checked_compilation_unit u
   let extend_signature ?(flags=Compiler.default_flags) ~base u = Compiler.append_unit_signature ~flags ~base u
 
+
+  module IntervalTree = Compiler.IntervalTree
+
+  type type_ = Compiler.type_
+  let pp_type_ = Compiler.pp_type_
+  type info = Compiler.info = { defined : Ast.Loc.t option; type_ : type_ option }
+  let pp_info = Compiler.pp_info
+  let hover = Compiler.hover
 end
 
 module Execute = struct
@@ -339,8 +350,8 @@ module RawOpaqueData = struct
       Format.fprintf fmt "@\n";
     end;
     Format.fprintf fmt "@[<hov 2>kind %s type.@]@\n@\n" name;
-    List.iter (fun (c,_) ->
-      Format.fprintf fmt "@[<hov 2>type %s %s.@]@\n" c name)
+    List.iter (fun (variant,(c,_)) ->
+      Format.fprintf fmt "@[<hov 2>external symbol %s : %s = \"%d\".@]@\n" c name variant)
       constants
     in
   { cin; cino; cout; isc; name = c },
@@ -348,13 +359,13 @@ module RawOpaqueData = struct
 
   let conversion_of_cdata (type a) ~name ?doc ?(constants=[]) ~compare ~pp cd =
     let module VM = Map.Make(struct type t = a let compare = compare end) in
-    let constants_map, values_map =
-      List.fold_right (fun (n,v) (cm,vm) ->
-        let c = ED.Global_symbols.declare_global_symbol n in
-        Util.Constants.Map.add c v cm, VM.add v c vm)
-      constants (Util.Constants.Map.empty,VM.empty) in
+    let constants_map, values_map, constants =
+      List.fold_right (fun (n,v) (cm,vm,cl) ->
+        let c, variant = ED.Global_symbols.declare_overloaded_global_symbol n in
+        Util.Constants.Map.add c v cm, VM.add v c vm,(variant,(n,v)) :: cl)
+      constants (Util.Constants.Map.empty,VM.empty,[]) in
     let values_map x = VM.find x values_map in
-    conversion_of_cdata ~name ?doc ~constants_map ~values_map ~constants ~pp cd
+    conversion_of_cdata ~name ?doc ~constants_map ~values_map ~constants:(List.rev constants) ~pp cd
 
   let declare { name; doc; pp; compare; hash; hconsed; constants; } =
     let cdata = Util.CData.declare {
@@ -485,23 +496,6 @@ module BuiltInData = struct
         aux (x :: acc) (gls :: extra) s xs
     in
       aux [] [] s l
-
-  let listC d =
-    let embed ~depth h c s l =
-      let module R = (val !r) in let open R in
-      let s, l, eg = map_acc (d.ContextualConversion.embed ~depth h c) s l in
-      s, list_to_lp_list l, eg in
-    let readback ~depth h c s t =
-      let module R = (val !r) in let open R in
-      map_acc (d.ContextualConversion.readback ~depth h c) s
-        (lp_list_to_list ~depth t)
-    in
-    let pp fmt l =
-      Format.fprintf fmt "[%a]" (Util.pplist d.pp ~boxed:true "; ") l in
-    { ContextualConversion.embed; readback;
-      ty = TyApp ("list",d.ty,[]);
-      pp;
-      pp_doc = (fun fmt () -> ()) }
   
   let list d =
     let embed ~depth s l =
@@ -524,10 +518,12 @@ end
 
 module Elpi = struct
 
-  type t = ED.uvar_body
+  type t = ED.uvar
 
-  let pp = Compiler.pp
+  let pp = Compiler.pp_uvar_body
   let show m = Format.asprintf "%a" pp m
+  let pp_raw = Compiler.pp_uvar_body_raw
+  let show_raw m = Format.asprintf "%a" pp_raw m
 
   let equal h1 h2  = h1 == h2
   let hash = ED.uvar_id 
@@ -539,7 +535,7 @@ module Elpi = struct
 
   let alloc_Elpi name state =
     let module R = (val !r) in
-    state, (ED.oref ED.dummy)
+    state, (ED.oref ~depth:0 ED.dummy)
 
   let make ?name state =
     match name with
@@ -559,7 +555,8 @@ end
 module RawData = struct
 
   type constant = Util.constant
-  type builtin = Util.constant
+  type builtin = ED.builtin_predicate = Cut | And | Impl | ImplBang | RImpl | Pi | Sigma | Eq | Match | Findall | Delay | Host of constant
+
   type term = ED.Term.term
   type view =
     (* Pure subterms *)
@@ -574,17 +571,17 @@ module RawData = struct
     | CData of RawOpaqueData.t                    (* external data *)
     (* Unassigned unification variables *)
     | UnifVar of Elpi.t * term list
-  [@@warning "-37"]
+  [@@warning "-37"]  
   
   let rec look ~depth t =
     let module R = (val !r) in let open R in
     match deref_head ~depth t with
     | ED.Term.Arg _ | ED.Term.AppArg _ -> assert false
-    | ED.Term.AppUVar(ub,0,args) -> UnifVar (ub,args)
-    | ED.Term.AppUVar(ub,lvl,args) -> look ~depth (R.expand_appuv ub ~depth ~lvl ~args)
-    | ED.Term.UVar(ub,lvl,ano) -> look ~depth (R.expand_uv ub ~depth ~lvl ~ano)
+    | ED.Term.AppUVar(ub,args) when ub.vardepth == 0 -> UnifVar (ub,args)
+    | ED.Term.AppUVar(ub,args) -> look ~depth (R.expand_appuv ub ~depth ~args)
+    | ED.Term.UVar(ub,ano) -> look ~depth (R.expand_uv ub ~depth ~ano)
     | ED.Term.Discard ->
-        let ub = ED.oref ED.dummy in
+        let ub = ED.oref ~depth:0 ED.dummy in
         UnifVar (ub,R.mkinterval 0 depth 0)
     | ED.Term.Lam _ as t ->
         begin match R.eta_contract_flex ~depth t with
@@ -594,7 +591,7 @@ module RawData = struct
     | x -> Obj.magic x (* HACK: view is a "subtype" of Term.term *)
 
   let kool = function
-    | UnifVar(ub,args) -> ED.Term.AppUVar(ub,0,args)
+    | UnifVar(ub,args) -> ED.Term.AppUVar(ub,args)
     | x -> Obj.magic x
   [@@ inline]
 
@@ -610,7 +607,7 @@ module RawData = struct
   let mkCons = ED.Term.mkCons
   let mkNil = ED.Term.mkNil
   let mkDiscard = ED.Term.mkDiscard
-  let mkBuiltin = ED.Term.mkBuiltin
+  let mkBuiltin c l = ED.Term.mkBuiltin c l
   let mkCData = ED.Term.mkCData
   let mkAppBoundL x l =
     if x < 0 then Util.anomaly "mkAppBoundL: got a global constant";
@@ -626,7 +623,7 @@ module RawData = struct
     if i < 0 then Util.anomaly "mkBound: got a global constant";
     mkConst i
 
-  let cmp_builtin i j = i - j
+  let cmp_builtin = ED.compare_builtin_predicate
 
   let mkAppMoreArgs ~depth hd args =
     let module R = (val !r) in let open R in
@@ -647,28 +644,21 @@ module RawData = struct
 
   module Constants = struct
 
-    let declare_global_symbol = ED.Global_symbols.declare_global_symbol
+    let declare_global_symbol ?variant x =
+      begin match variant with
+      | Some n -> if n < 1 then Util.error "declare_global_symbol: variants are >= 0"
+      | None -> () end;
+      ED.Global_symbols.declare_global_symbol ?variant x
 
     let show c = Util.Constants.show c
 
-    let eqc    = ED.Global_symbols.eqc
     let orc    = ED.Global_symbols.orc
-    let andc   = ED.Global_symbols.andc
-    let rimplc = ED.Global_symbols.rimplc
-    let pic    = ED.Global_symbols.pic
-    let sigmac = ED.Global_symbols.sigmac
-    let implc  = ED.Global_symbols.implc
-    let cutc   = ED.Global_symbols.cutc
-    let ctypec = ED.Global_symbols.ctypec
-    let spillc = ED.Global_symbols.spillc
 
     module Map = Util.Constants.Map
     module Set = Util.Constants.Set
 
   end
 
-  let of_term x = x
-
   let of_hyp x = x
   let of_hyps x = x
 
@@ -677,19 +667,21 @@ module RawData = struct
     hsrc : term
   }
   type hyps = hyp list
-
+  type blockers = Elpi.t list
   type suspended_goal = ED.suspended_goal = {
     context : hyps;
-    goal : int * term
+    goal : int * term;
+    blockers : blockers;
   }
 
   let constraints l =
     let module R = (val !r) in let open R in
-    Util.map_filter (fun x -> get_suspended_goal x.ED.kind) l
+    Util.map_filter (fun (x : ED.stuck_goal) -> get_suspended_goal x.ED.blockers x.ED.kind) l
   let no_constraints = []
 
   let mkUnifVar ub ~args state =
-    ED.Term.mkAppUVar ub 0 args
+    if args = [] then ED.Term.mkUVar ub 0
+    else ED.Term.mkAppUVar ub args
 
   type Conversion.extra_goal +=
   | RawGoal = ED.Conversion.RawGoal
@@ -768,7 +760,7 @@ module FlexibleData = struct
       let pp k uv _ () =
            Format.fprintf fmt "@[<h>%a@ <-> %a@]@ " T.pp k Elpi.pp uv
         in
-      Format.fprintf fmt "@[<v>";
+      Format.fprintf fmt "@[<hov>";
       fold pp m ();
       Format.fprintf fmt "@]"
     ;;
@@ -807,14 +799,34 @@ module AlgebraicData = struct
   type name = string
   type doc = string
 
-  let declare x =
+  let allocate_constructors x =
     let module R = (val !r) in
-    ED.BuiltInPredicate.ADT.adt
+    ED.BuiltInPredicate.ADT.allocate_constructors
       ~look:R.deref_head
       ~mkinterval:R.mkinterval
       ~mkConst:R.mkConst
       ~alloc:FlexibleData.Elpi.make
       ~mkUnifVar:RawData.mkUnifVar x
+
+  let declare_allocated alloc x =
+    let module R = (val !r) in
+    ED.BuiltInPredicate.ADT.declare_allocated
+      ~look:R.deref_head
+      ~mkinterval:R.mkinterval
+      ~mkConst:R.mkConst
+      ~alloc:FlexibleData.Elpi.make
+      ~mkUnifVar:RawData.mkUnifVar alloc x
+
+  let declare x =
+    begin match x.ED.BuiltInPredicate.ADT.ty with
+    | TyApp(s,_,_) -> Util.error ("Declaration of " ^ s ^ " requires allocate_constructors + declare_allocated")
+    | TyName _ -> ()
+    end;
+    let x = ED.BuiltInPredicate.ADT.Decl x in
+    let alloc = allocate_constructors x in
+    declare_allocated alloc x
+
+       
 end
 
 module BuiltInPredicate = struct
@@ -894,7 +906,7 @@ module BuiltInPredicate = struct
 
   let beta ~depth t args =
     let module R = (val !r) in let open R in
-    deref_appuv ~from:depth ~to_:depth ?avoid:None args t
+    deref_apparg ~from:depth ~to_:depth ?avoid:None t args
 
   module HOAdaptors = struct
 
@@ -904,7 +916,7 @@ module BuiltInPredicate = struct
     type ('a,'b,'c) pred3 = Data.term * 'a Conversion.t * 'b Conversion.t * 'c Conversion.t
     type ('a,'b) pred3a = Data.term * 'a Conversion.t * 'b Conversion.t
 
-    let pred1_ty x = Conversion.TyApp("->",x.Conversion.ty,[Conversion.TyName"prop"])
+    let pred1_ty x = Conversion.TyApp("->",x.Conversion.ty,[Conversion.TyName"(func)"])
     let pred1 x = { Conversion.ty = pred1_ty x; readback = (fun ~depth state e -> state,(e,x),[]); embed = (fun ~depth state (x,_) -> state,x,[]); pp = (fun fmt (x,_) -> Format.fprintf fmt "<pred1>"); pp_doc = (fun fmt () -> ()) }
     let pred2_ty x y = Conversion.(TyApp("->",x.Conversion.ty,[TyApp("->",y.Conversion.ty,[Conversion.TyName"prop"])]))
     let pred2 x y = { Conversion.ty = pred2_ty x y; readback = (fun ~depth state e -> state,(e,x,y),[]); embed = (fun ~depth state (x,_,_) -> state,x,[]); pp = (fun fmt (x,_,_) -> Format.fprintf fmt "<pred2>"); pp_doc = (fun fmt () -> ()) }
@@ -1087,17 +1099,19 @@ module Calc = struct
   }
 
   let compile_operation_declaration { symbol; infix; args; code } =
-    let c = ED.Global_symbols.declare_global_symbol symbol in
     let ty_decl args =
-     if infix then
-       Printf.sprintf "type (%s) %s." symbol (String.concat " -> " args)
-     else
-       Printf.sprintf "type %s %s." symbol (String.concat " -> " args) in
-     c, { ED.CalcHooks.ty_decl = List.map ty_decl args |> String.concat "\n"; code }
+      let c, variant = ED.Global_symbols.declare_overloaded_global_symbol symbol in
+      let ty_decl = if infix then
+        Printf.sprintf "external symbol (%s) : %s = \"%d\". " symbol (String.concat " -> " args) variant
+      else
+        Printf.sprintf "external symbol %s : %s = \"%d\"." symbol (String.concat " -> " args) variant in
+      c, { ED.CalcHooks.ty_decl = ty_decl; code }
+    in
+    List.map ty_decl args
 
    let register ~descriptor d =
      let e = compile_operation_declaration d in
-     descriptor := e :: !descriptor
+     descriptor := e @ !descriptor
     
   let register_eval n (symbol,tys) code =
     let infix, n = n < 0, abs n in
@@ -1256,7 +1270,7 @@ module Calc = struct
     [
     LPDoc " -- Evaluation --";
 
-    LPCode "pred (is) o:A, i:A.";
+    LPCode ":functional pred (is) o:A, i:A.";
     LPCode "X is Y :- calc Y X.";
   
     MLCode(Pred("calc",
@@ -1289,7 +1303,6 @@ module Utils = struct
     hmove ~from ~to_ ?avoid:None t
 
   let beta = BuiltInPredicate.beta
-
   let error = Util.error
   let type_error = Util.type_error
   let anomaly = Util.anomaly
@@ -1312,7 +1325,7 @@ module Utils = struct
       | Data.Lam t ->
           let s = "x" ^ string_of_int d in
           let ctx = Util.IntMap.add d (Term.mkCon buggy_loc s) ctx in
-          Term.mkLam buggy_loc s None (aux (d+1) ctx t)
+          Term.mkLam buggy_loc s buggy_loc None (aux (d+1) ctx t)
       | Data.App(c,x,xs) ->
           let c = aux d ctx (R.mkConst c) in
           let x = aux d ctx x in
@@ -1325,7 +1338,7 @@ module Utils = struct
           Term.mkSeq [hd;tl]
       | Data.Nil -> Term.mkNil buggy_loc
       | Data.Builtin(c,xs) ->
-          let c = Term.mkCon buggy_loc (show c) in
+          let c = Term.mkCon buggy_loc (ED.show_builtin_predicate (fun ?table -> show) c) in
           let xs = List.map (aux d ctx) xs in
           Term.mkApp loc (c :: xs)
       | Data.CData x -> Term.mkC buggy_loc x
@@ -1345,7 +1358,7 @@ module Utils = struct
       Clause.loc = loc;
       attributes;
       body = aux depth Util.IntMap.empty term;
-      needs_spilling = ();
+      needs_spilling = ()
     }]
 
   let term_to_raw_term s p ?ctx ~depth t =
@@ -1355,6 +1368,7 @@ module Utils = struct
   let map_acc = BuiltInData.map_acc
 
   module type Show = Util.Show
+  module type ShowKey = Util.ShowKey
   module type Show1 = Util.Show1
   module Map = Util.Map
   module Set = Util.Set
@@ -1363,6 +1377,76 @@ module Utils = struct
 
   let version_parser = Util.version_parser
 
+  let error_cmp_flex ~depth t1 t2 = error "cmp_term on non-ground terms"
+
+let rec cmp_term ~depth t1 t2 =
+  let open RawData in
+  match look ~depth t1, look ~depth t2 with
+  | Nil, Nil -> 0
+  | Nil, (Cons _ | Const _ | App _ | Lam _ | Builtin _ | CData _ | UnifVar _) -> -1
+
+  | Cons(x,xs), Cons(y,ys) ->
+      let cmp_x = cmp_term ~depth x y in
+      if cmp_x == 0 then cmp_term ~depth xs ys
+      else cmp_x
+  | Cons _, (Const _ | App _ | Lam _ | Builtin _ | CData _ | UnifVar _) -> -1
+  | Cons _, Nil -> 1
+
+  | Const c1, Const c2 -> c1 - c2
+  | Const _, (App _ | Lam _ | Builtin _ | CData _ | UnifVar _) -> -1
+  | Const _, (Cons _ | Nil) -> 1
+
+  | Lam t1, Lam t2 -> cmp_term ~depth:(depth+1) t1 t2
+  | Lam _, (App _ | Builtin _ | CData _ | UnifVar _) -> -1
+  | Lam _, (Const _ | Cons _ | Nil) -> 1
+
+  | App(c1,x,xs), App(c2,y,ys) ->
+      let cmp_c1 = c1 - c2 in
+      if cmp_c1 == 0 then
+        let cmp_x = cmp_term ~depth x y in
+        if cmp_x == 0 then cmp_terms ~depth xs ys else cmp_x
+      else cmp_c1
+  | App _, (Builtin _ | CData _ | UnifVar _) -> -1
+  | App _, (Lam _ | Const _ | Cons _ | Nil) -> 1
+
+  | Builtin(c1,xs), Builtin(c2,ys) ->
+      let cmp_c1 = cmp_builtin c1 c2 in
+      if cmp_c1 == 0 then cmp_terms ~depth xs ys else cmp_c1
+  | Builtin _, (CData _ | UnifVar _) -> -1
+  | Builtin _, (App _ | Lam _ | Const _ | Cons _ | Nil) -> 1
+
+  | CData d1, CData d2 -> RawOpaqueData.compare d1 d2
+  | CData _, UnifVar _ -> -1
+  | CData _, (Builtin _ | App _ | Lam _ | Const _ | Cons _ | Nil) -> 1
+
+  | UnifVar(b1,xs), UnifVar(b2,ys) ->
+      if FlexibleData.Elpi.equal b1 b2 then
+        if cmp_terms ~depth xs ys == 0 then 0
+        else error_cmp_flex ~depth t1 t2
+      else error_cmp_flex ~depth t1 t2
+  | UnifVar _, (CData _ | Builtin _ | App _ | Lam _ | Const _ | Cons _ | Nil) -> 1
+
+and cmp_terms ~depth l1 l2 =
+  match l1, l2 with
+  | [], [] -> 0
+  | [], _ :: _ -> -1
+  | _ :: _, [] -> 1
+  | x :: xs, y :: ys ->
+      let cmp_x = cmp_term ~depth x y in
+      if cmp_x == 0 then cmp_terms ~depth xs ys else cmp_x
+
+let rec check_ground ~depth t =
+  let open RawData in
+  match look ~depth t with
+  | Nil | Const _ | CData _ -> ()
+  | Lam t -> check_ground ~depth:(depth + 1) t
+  | Cons(x,xs) -> check_ground ~depth x; check_ground ~depth xs
+  | Builtin(_,l) -> List.iter (check_ground ~depth) l
+  | App(_,x,xs) -> check_ground ~depth x; List.iter (check_ground ~depth) xs
+  | UnifVar _ -> raise BuiltInPredicate.No_clause
+
+
+
 end
 
 module RawPp = struct
