@@ -329,6 +329,31 @@ let match_file ~log file adjust reference =
     | _ -> false
   end
 
+let match_json_file ~log file adjust reference =
+  let file = adjust file in
+  let dig1 = Digest.to_hex (Digest.file file) in
+  let dig2 = Digest.(to_hex (file reference)) in
+  if dig1 = dig2 then begin
+    Util.write log (Printf.sprintf "Digest of %s and %s is %s\n" file reference dig2);
+    true
+  end else
+    let yreference = Yojson.Safe.from_file reference in
+    let yfile = Yojson.Safe.from_file file in
+    Yojson.Safe.equal yreference yfile ||
+  begin
+    Util.write log (Printf.sprintf "Diffing %s against %s\n" file reference);
+      match
+      Util.exec ~timeout:5.0 ~env:(Unix.environment ())
+        ~log ~close_output:false
+        ~executable:"diff" ~args:["-u";"--strip-trailing-cr";reference;file] ()
+    with
+    | Util.Exit(0,_,_) -> true
+    | Util.Exit(n,_,_) ->
+      Util.write log (Printf.sprintf "Exit code: %d\n" n);
+      Util.write log (Printf.sprintf "Promotion: cp %s %s\n" file reference);
+      false
+    | _ -> false
+  end
 
 module Elpi = struct
 
@@ -367,10 +392,18 @@ let read_tctime input_line =
   let time = ref 0.0 in
   try while true do
     let l = input_line () in
-    if Str.(string_match (regexp "^Typechecking time:") l 0) then
-      try time := float_of_string (String.sub l 18 (String.length l - 18))
+    begin
+      if Str.(string_match (regexp "^Typechecking time:") l 0) then
+      try time := !time +. float_of_string (String.sub l 18 (String.length l - 18))
       with _ -> ()
-  done; !time
+    end;
+    begin
+      if Str.(string_match (regexp "^Determinacy checking time:") l 0) then
+        try time := !time +. float_of_string (String.sub l 26 (String.length l - 26))
+      with _ -> ()
+    end;
+    done;
+    !time
   with End_of_file -> !time
 
 let () = Runner.declare
@@ -395,6 +428,7 @@ let () = Runner.declare
       if outside_llam then "-delay-problems-outside-pattern-fragment"::args
       else args in
     Util.write log (Printf.sprintf "args: %s\n" (String.concat " " args));
+    Util.write log (Printf.sprintf "\ndebug with: dune exec elpi -- %s -trace-on -trace-at run 1 99999 -trace-only user\n\n" (String.concat " " args));
 
     let rc =
       let outcome = Util.exec ~timeout ~timetool ~close_output:false ?input ~executable ~env ~log ~args () in
@@ -477,7 +511,7 @@ module ElpiTraceElab = struct
         match outcome, outcomey, test.Test.expectation with
         | Util.Exit(0,walltime,mem), Some(Util.Exit(0,_,_)), Test.SuccessOutputTxt f when f (Util.read_lines (snd log)) -> 
             Runner.Success { walltime; typechecking; execution; mem }
-        | Util.Exit(0,walltime,mem), Some(Util.Exit(0,_,_)), Test.SuccessOutputFile { sample; adjust; reference } when match_file ~log sample adjust (sources^"/"^reference) ->
+        | Util.Exit(0,walltime,mem), Some(Util.Exit(0,_,_)), Test.SuccessOutputFile { sample; adjust; reference } when match_json_file ~log sample adjust (sources^"/"^reference) ->
             Runner.Success { walltime; typechecking; execution; mem }
         | Util.Exit(0,walltime,mem), Some(Util.Exit(0,_,_)), Test.SuccessOutputFile { sample; adjust; reference } when promote ->
             FileUtil.cp [adjust sample] (sources^"/"^reference);
