@@ -14,10 +14,10 @@ module Printer : sig
     executables:string list -> seed:int -> timeout:float -> unit
     
   val print_summary :
-    total:int -> ok:int -> ko:int -> skipped:int -> unit
+    total:int -> ok:int -> ko_list:string list -> skipped:int -> timeout:string list -> unit
 
   val print_log :
-    fname:string -> unit
+    ln_nb:int -> fname:string -> unit
 
 end = struct
 open ANSITerminal
@@ -46,30 +46,47 @@ let print_header ~executables ~seed ~timeout =
   printf [blue] "------------------------------------------------------------------\n";
 ;;
 
-let print_summary ~total ~ok ~ko ~skipped =
+let print_summary ~total ~ok ~ko_list ~skipped ~timeout =
   printf [blue] "------------------------------------------------------------------\n";
-  printf [blue] "Tests: "; printf [] "%d\n" total;
-  printf [blue] "Passed: "; printf [] "%d\n" ok;
-  printf [blue] "Failed: "; printf [] "%d\n" ko;
-  printf [blue] "Skipped: "; printf [] "%d\n" skipped;
+  let print_stat ?(to_print=false) k v =
+    if to_print || v <> 0 then (printf [blue] "%s: " k; printf [] "%d\n" v)
+  in
+  print_stat ~to_print:true "Tests" total;
+  print_stat ~to_print:true "Passed" ok;
+  print_stat ~to_print:true "Failed" (List.length ko_list);
+  print_stat "Skipped" skipped;
+  print_stat "Timeout" (List.length timeout);
+  if ko_list <> [] then 
+    printf [red] "Rerun failed: make tests ONLY=\"'^\\(%s\\)'\"\n" (String.concat "\\|" ko_list);
+  if timeout <> [] then 
+    printf [red] "Rerun timeout: make tests ONLY=\"'^\\(%s\\)'\"\n" (String.concat "\\|" timeout)
 ;;
 
-let print_file fname =
+let print_file ~ln_nb fname =
+  let pos = ref 0 in
+  let arr = Array.init (max 0 ln_nb) (fun _ -> fun () -> ()) in
+  let print_s s () = printf [] "%s\n" s in
+  let print_aux =
+    if ln_nb = 0 then fun _ -> () else
+    if ln_nb > ~-1 then 
+      fun s -> (arr.(!pos) <- print_s s; pos := (!pos + 1) mod ln_nb;)
+    else fun s -> printf [] "%s\n" s in
   try
     let ic = open_in fname in
     while true do
       let s = input_line ic in
-      printf [] "%s\n" s
+      print_aux s
     done
   with
-  | End_of_file -> ()
+  | End_of_file -> for i = 0 to ln_nb - 1 do arr.((i + !pos) mod ln_nb) (); done
   | e -> printf [red] "Error reading %s: %s\n" fname (Printexc.to_string e)
 
-let print_log ~fname =
+
+let print_log ~ln_nb ~fname =
   printf [red] "------------------------------------------------------------------\n";
   printf [blue] "Log of the first failure: "; printf [] "%s\n" fname;
   printf [red] "------------------------------------------------------------------\n";
-  print_file fname;
+  print_file ~ln_nb fname;
   printf [red] "------------------------------------------------------------------\n";
   printf [blue] "End log of the first failure: "; printf [] "%s\n" fname;
   printf [red] "------------------------------------------------------------------\n";
@@ -80,7 +97,7 @@ let aNSITerminal_move_bol () =
   if Sys.win32 then ANSITerminal.printf [] "\n%!"
   else ANSITerminal.move_bol ()
 
-let run timeout _seed sources promote env { Runner.run; test; executable }  =
+let run stop_on_first_error timeout _seed sources promote env { Runner.run; test; executable }  =
 
   let { Test.name; description; _ } = test in
   let print = Printer.print ~executable:(Filename.basename executable) ~name ~description ~promote in
@@ -123,7 +140,9 @@ let print_csv plot results =
           end
       | None -> ());
   close_out oc;
-  if Sys.command "which lua5.1" = 0 && Sys.command "which gnuplot" = 0 then begin
+  let command_exists cmd =
+    Sys.command ("command -v " ^ cmd ^ " 2> /dev/null || exit 1 ") = 0 in
+  if command_exists "lua5.1" && command_exists "gnuplot" then begin
     ignore(Sys.command (plot ^ " data.csv"));
     ignore(Sys.command "gnuplot data.csv.plot")
   end
@@ -136,7 +155,13 @@ let rec find_map f = function
       | Some y -> y
       | None -> find_map f xs
 
-let main sources plot timeout promote executables namef catskip timetool seed =
+let rec map_stop_on_error stop_on_first_error f = function
+  | [] -> []
+  | x :: xs -> 
+    match f x with
+    | Some Runner.{rc=Failure _} as err when stop_on_first_error -> [err]
+    | e -> e :: map_stop_on_error stop_on_first_error f xs
+let main ln_nb sources plot timeout promote executables namef catskip timetool seed stop_on_first_error =
   Random.init seed;
   let filter_name =
     let rex = Str.regexp (".*"^namef) in
@@ -149,26 +174,27 @@ let main sources plot timeout promote executables namef catskip timetool seed =
     tests |> List.map (Suite.Runner.jobs ~timetool ~executables ~promote)
           |> List.concat in
   let results =
-    List.map (run timeout seed sources promote env) jobs in
-  let total, ok, ko, skipped =
-    let skip, rest =
-      List.partition (function None -> true | _ -> false) results in
-    let ok, ko =
-      List.partition (function
-        | Some { Runner.rc = Runner.Success _; _ } -> true
-        | _ -> false) rest in
-    List.(length jobs, length ok, length ko, length skip) in
-  Printer.print_summary ~total ~ok ~ko ~skipped;
+    map_stop_on_error stop_on_first_error (run stop_on_first_error timeout seed sources promote env) jobs in
+  let total, ok, ko_list, skipped, timeout =
+    let rec part total ok ko_list skipped timeout = function
+      | [] -> (total, ok, List.rev ko_list, skipped, List.rev timeout)
+      | Some {Runner.rc = Success _; _} :: l -> part (total+1) (ok+1) ko_list skipped timeout l
+      | Some {rc = Promote _; _} :: l ->        part (total+1) (ok+1) ko_list skipped timeout l
+      | Some {rc = Failure _; test} :: l ->        part (total+1) ok (test.name :: ko_list) skipped timeout l
+      | None :: l ->                            part (total+1) ok ko_list (skipped+1) timeout l
+      | Some {rc = Timeout _; test} :: l ->        part (total+1) ok ko_list skipped (test.name :: timeout) l
+    in part 0 0 [] 0 [] results in
+  Printer.print_summary ~total ~ok ~ko_list ~skipped ~timeout;
   begin try
     let log_first_failure =
       results |> find_map (function
         | Some { Runner.rc = Runner.Failure _; log; _ } -> Some log
         | _ -> None) in
-    Printer.print_log ~fname:log_first_failure
+    Printer.print_log ~ln_nb:ln_nb ~fname:log_first_failure
   with Not_found -> ()
   end;
   if List.length executables > 1 then print_csv plot results;
-  if ko = 0 then exit 0 else exit 1
+  if ko_list = [] then exit 0 else exit 1
 
 open Cmdliner
 
@@ -215,6 +241,13 @@ let promote =
   let doc = "Promotes the tests (if failing)" in
   Arg.(value & opt bool false & info ["promote"] ~docv:"PATH" ~doc)
 
+let ln_nb =
+  let doc = "Sets the maximum number of lines to print for failing test (-1 means no max)" in
+  Arg.(value & opt int ~-1 & info ["ln_nb"] ~docv:"INT" ~doc)
+let stop_on_first_error = 
+  let doc = "Stops test execution on first error" in
+  Arg.(value & opt bool false & info ["stop-on-first-error"] ~docv:"PATH" ~doc)
+
 let info =
   let doc = "run the test suite" in
   let tests = Test.names ()
@@ -226,5 +259,5 @@ let info =
 ;;
 
 let () =
-  (Term.exit @@ Term.eval (Term.(const main $ src $ plot $ timeout $ promote $ runners $ namef $ catskip $ mem $ seed),info)) [@ warning "-A"]
+  (Term.exit @@ Term.eval (Term.(const main $ ln_nb $ src $ plot $ timeout $ promote $ runners $ namef $ catskip $ mem $ seed $ stop_on_first_error),info)) [@ warning "-A"]
   (* ocaml >= 4.08 | exit @@ Cmd.eval (Cmd.v info Term.(const main $ src $ plot $ timeout $ runners $ namef $ catskip $ mem $ seed)) *)
